@@ -2,14 +2,16 @@ from biblemate.core.systems import *
 from biblemate.ui.prompts import getInput
 from biblemate.ui.info import get_banner
 from biblemate.ui.selection_dialog import TerminalModeDialogs
-from biblemate import config, AGENTMAKE_CONFIG, OLLAMA_NOT_FOUND, fix_string, BIBLEMATEDATA
-from biblemate.uba.bible import BibleVectorDatabase
+from biblemate import config, AGENTMAKE_CONFIG, BIBLEMATEDATA, fix_string, write_user_config
+from biblemate.uba.api import run_uba_api, DEFAULT_MODULES
 from pathlib import Path
-import asyncio, re, os, subprocess, click, shutil, pprint, argparse, json
+import urllib.parse
+import asyncio, re, os, subprocess, click, gdown, pprint, argparse, json, zipfile
 from copy import deepcopy
 from alive_progress import alive_bar
 from fastmcp import Client
-from agentmake import agentmake, getOpenCommand, getDictionaryOutput, edit_configurations, readTextFile, writeTextFile, getCurrentDateTime, AGENTMAKE_USER_DIR, USER_OS, DEVELOPER_MODE, DEFAULT_AI_BACKEND
+from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
+from agentmake import agentmake, getOpenCommand, getDictionaryOutput, edit_file, edit_configurations, readTextFile, writeTextFile, getCurrentDateTime, AGENTMAKE_USER_DIR, USER_OS, DEVELOPER_MODE, DEFAULT_AI_BACKEND
 from agentmake.utils.handle_text import set_log_file_max_lines
 from rich.console import Console
 from rich.markdown import Markdown
@@ -25,27 +27,9 @@ if not os.path.isdir(log_path):
 log_file = os.path.join(log_path, "requests")
 set_log_file_max_lines(log_file, config.max_log_lines)
 
-# bible data
-builtin_bible_data = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data", "bibles")
-user_bible_data = os.path.join(BIBLEMATEDATA, "bibles")
-if not os.path.isdir(user_bible_data):
-    Path(user_bible_data).mkdir(parents=True, exist_ok=True)
-user_bible = os.path.join(user_bible_data, f"{config.default_bible}.bible")
-if not os.path.isfile(user_bible):
-    print("# Copying bible data ...")
-    shutil.copyfile(os.path.join(builtin_bible_data, f"{config.default_bible}.bible"), user_bible)
-if os.path.isfile(user_bible) and os.path.getsize(user_bible) < 380000000:
-    if shutil.which("ollama"):
-        print("# Setting up a bible vector database to support semantic search with BibleMate AI. please kindly wait until it is finished ...")
-        db = BibleVectorDatabase(user_bible)
-        db.add_vectors()
-        db.clean_up()
-        del db
-    else:
-        print(OLLAMA_NOT_FOUND)
-
 # AI backend
-parser = argparse.ArgumentParser(description = """BibleMate AI CLI options""")
+version = readTextFile(os.path.join(os.path.dirname(os.path.realpath(__file__)), "version.txt"))
+parser = argparse.ArgumentParser(description = f"""BibleMate AI {version} CLI options""")
 parser.add_argument("-b", "--backend", action="store", dest="backend", help="AI backend; overrides the default backend temporarily.")
 parser.add_argument("-mcp", "--mcp", action="store", dest="mcp", help=f"specify a custom MCP server to use, e.g. 'http://127.0.0.1:{config.mcp_port}/mcp/'; applicable to command `biblemate` only")
 parser.add_argument("-p", "--port", action="store", dest="port", help=f"specify a port for the MCP server to use, e.g. {config.mcp_port}; applicable to command `biblematemcp` only")
@@ -141,10 +125,15 @@ Get a static text-based response directly from a text-based AI model without usi
     
     return tools, tools_schema, master_available_tools, available_tools, tool_descriptions, prompts, prompts_schema, resources, templates
 
-def backup_conversation(console, messages, master_plan):
+def backup_conversation(messages, master_plan, console=None):
     """Backs up the current conversation to the user's directory."""
-    timestamp = getCurrentDateTime()
-    storagePath = os.path.join(AGENTMAKE_USER_DIR, "biblemate", "chats", timestamp)
+    # determine storage path
+    if console:
+        timestamp = getCurrentDateTime()
+        storagePath = os.path.join(AGENTMAKE_USER_DIR, "biblemate", "chats", timestamp)
+    else:
+        storagePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), "temp")
+    # create directory if not exists
     if not os.path.isdir(storagePath):
         Path(storagePath).mkdir(parents=True, exist_ok=True)
     # Save full conversation
@@ -157,27 +146,13 @@ def backup_conversation(console, messages, master_plan):
     markdown_text = "\n\n".join(["```"+i["role"]+"\n"+i["content"]+"\n```" for i in messages if i.get("role", "") in ("user", "assistant")])
     writeTextFile(markdown_file, markdown_text)
     # Save html
-    html_file = os.path.join(storagePath, "conversation.html")
-    console.save_html(html_file, inline_styles=True, theme=MONOKAI)
+    if console:
+        html_file = os.path.join(storagePath, "conversation.html")
+        console.save_html(html_file, inline_styles=True, theme=MONOKAI)
     # Inform users of the backup location
-    print(f"Conversation backup saved to {storagePath}")
-    print(f"Report saved to {html_file}\n")
-
-def write_user_config():
-    """Writes the current configuration to the user's config file."""
-    config_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "config.py")
-    configurations = f"""agent_mode={config.agent_mode}
-prompt_engineering={config.prompt_engineering}
-max_steps={config.max_steps}
-lite={config.lite}
-hide_tools_order={config.hide_tools_order}
-default_bible="{config.default_bible}"
-max_semantic_matches={config.max_semantic_matches}
-max_log_lines={config.max_log_lines}
-mcp_port={config.mcp_port}
-embedding_model="{config.embedding_model}"
-disabled_tools={pprint.pformat(config.disabled_tools)}"""
-    writeTextFile(config_file, configurations)
+    if console:
+        print(f"Conversation backup saved to {storagePath}")
+        print(f"Report saved to {html_file}\n")
 
 async def main_async():
 
@@ -201,6 +176,11 @@ async def main_async():
 
     async with client:
         tools, tools_schema, master_available_tools, available_tools, tool_descriptions, prompts, prompts_schema, resources, templates = await initialize_app(client)
+        resource_suggestions = json.loads(run_uba_api(".resources"))
+        resource_suggestions = resource_suggestions["bibleListAbb"]+resource_suggestions["commentaryListAbb"]+resource_suggestions["encyclopediaListAbb"]+resource_suggestions["lexiconList"]
+        abbr = BibleBooks.abbrev["eng"]
+        resource_suggestions += [abbr[str(book)][0] for book in range(1,67)]
+
         write_user_config() # remove the temporary `config.backend`
         
         available_tools_pattern = "|".join(available_tools)
@@ -259,19 +239,17 @@ async def main_async():
             elif APP_START:
                 print()
                 APP_START = False
-                while True:
-                    try:
-                        agentmake("Hello!", system=DEFAULT_SYSTEM)
-                        break
-                    except Exception as e:
-                        print("Connection failed! Please ensure that you have a stable internet connection and that my AI backend and model are properly configured.")
-                        print("Viist https://github.com/eliranwong/agentmake#supported-backends for help about the backend configuration.\n")
-                        if click.confirm("Do you want to configure my AI backend and model now?", default=True):
-                            edit_configurations()
-                            console.rule()
-                            console.print("Restart to make the changes in the backend effective!", justify="center")
-                            console.rule()
-                            exit()
+                try:
+                    agentmake("Hello!", system=DEFAULT_SYSTEM)
+                except Exception as e:
+                    print("Connection failed! Please ensure that you have a stable internet connection and that my AI backend and model are properly configured.")
+                    print("Viist https://github.com/eliranwong/agentmake#supported-backends for help about the backend configuration.\n")
+                    if click.confirm("Do you want to configure my AI backend and model now?", default=True):
+                        edit_configurations()
+                        console.rule()
+                        console.print("Restart to make the changes in the backend effective!", justify="center")
+                        console.rule()
+                        exit()
             # Original user request
             # note: `python3 -m rich.emoji` for checking emoji
             console.print("Enter your request :smiley: :" if len(messages) == len(DEFAULT_MESSAGES) else "Enter a follow-up request :flexed_biceps: :")
@@ -282,17 +260,19 @@ async def main_async():
                 ".mode": "change AI mode",
                 ".tools": "list available tools",
                 ".plans": "list available plans",
-                ".resources": "list available resources",
+                ".resources": "list UniqueBible resources",
                 ".promptengineer": "toggle auto prompt engineering",
                 ".lite": "toggle lite context",
                 ".steps": "configure the maximum number of steps",
                 ".matches": "configure the maximum number of semantic matches",
+                ".edit": "load the current conversation",
                 ".backup": "backup conversation",
                 ".load": "load a saved conversation",
                 ".open": "open a file or directory",
+                ".download": "download data files",
                 ".help": "help page",
             }
-            input_suggestions = list(action_list.keys())+["@ ", "@@ "]+[f"@{t} " for t in available_tools]+[f"{p} " for p in prompt_list]+[f"//{r}" for r in resources.keys()]+template_list # "" is for generating ideas
+            input_suggestions = list(action_list.keys())+["@ ", "@@ "]+[f"@{t} " for t in available_tools]+[f"{p} " for p in prompt_list]+[f"//{r}" for r in resources.keys()]+template_list+resource_suggestions # "" is for generating ideas
             user_request = await getInput("> ", input_suggestions)
             while not user_request.strip():
                 # Generate ideas for `prompts to try`
@@ -323,36 +303,48 @@ async def main_async():
                     else:
                         display_content = resource_text
                     resource_description = resources.get(resource, "")
-                    console.print(Markdown(f"## Resource: `{resource.capitalize()}`\n\n{resource_description}\n\n{display_content}"))
+                    console.print(Markdown(f"## Information about `{resource}`: `{resource.capitalize()}`\n\n{resource_description}\n\n{display_content}"))
                     console.rule()
                 continue
 
             # run templates
             if re.search(template_pattern, user_request):
+                user_request = urllib.parse.quote(user_request)
+                if user_request[2:].count("/") == 1:
+                    keywords = DEFAULT_MODULES
+                    keyword, entry = user_request[2:].split("/")
+                    if module := keywords.get(keyword, ""):
+                        user_request = f"//{keyword}/{module}/{entry}"
+                        if user_request.count("/") > 4:
+                            user_request = re.sub("^(//.*?/.*?/)(.*?)$", r"\1"+r"\2".replace("/", "「」"), user_request)
+                    elif user_request.count("/") > 3:
+                        user_request = re.sub("^(//.*?/)(.*?)$", r"\1"+r"\2".replace("/", "「」"), user_request)
                 try:
                     uri = re.sub("^(.*?)/", r"\1://", user_request[2:])
                     resource_content = await client.read_resource(uri)
                     resource_content = resource_content[0].text
                     while resource_content.startswith("[") and resource_content.endswith("]"):
                         options = json.loads(resource_content)
-                        select = await dialogs.getMultipleSelection(
+                        select = await dialogs.getValidOptions(
                             options=options,
                             title="Multiple Matches",
                             text="Select one of them to continue:"
                         )
                         if select:
-                            resource_content = await client.read_resource(re.sub("^(.*?/)[^/]*?$", r"\1", uri)+select)
-                            resource_content = resource_content[0].text
+                            if keyword == "name":
+                                resource_content = select
+                            else:
+                                resource_content = await client.read_resource(re.sub("^(.*?/)[^/]*?$", r"\1", uri)+urllib.parse.quote(select.replace("/", "「」")))
+                                resource_content = resource_content[0].text
                         else:
                             resource_content = "Cancelled by user."
                     if resource_content:
                         messages += [
-                            {"role": "user", "content": f"Retrieve resource from:\n\n{uri}"},
+                            {"role": "user", "content": f"Retrieve content from:\n\n{uri}"},
                             {"role": "assistant", "content": resource_content},
                         ]
                         console.rule()
                         console.print(Markdown(resource_content))
-                        console.rule()
                     continue
                 except Exception as e: # invalid uri
                     print(f"Error: {e}\n")
@@ -366,15 +358,30 @@ async def main_async():
                 cmd = f'''{getOpenCommand()} "{file_path}"'''
                 subprocess.Popen(cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                 continue
-            elif user_request.startswith(".load") and re.search('''.py['" ]*?$''', user_request) and os.path.isfile(os.path.expanduser(re.sub('''^['" ]*?([^'" ].+?)['" ]*?$''', r"\1", user_request[6:]))):
+            elif user_request.startswith(".load") and os.path.exists(os.path.expanduser(re.sub('''^['" ]*?([^'" ].+?)['" ]*?$''', r"\1", user_request[6:]))):
+                load_path = os.path.expanduser(re.sub('''^['" ]*?([^'" ].+?)['" ]*?$''', r"\1", user_request[6:]))
                 try:
-                    file_path = os.path.expanduser(re.sub('''^['" ]*?([^'" ].+?)['" ]*?$''', r"\1", user_request[6:]))
-                    backup_conversation(console, messages, master_plan)
+                    # load conversation
+                    if os.path.isfile(load_path):
+                        file_path = load_path
+                    elif os.path.isdir(load_path) and os.path.isfile(os.path.join(load_path, "conversation.py")) and os.isfile(os.path.join(load_path, "master_plan.md")):
+                        file_path = os.path.join(load_path, "conversation.py")
+                    else:
+                        print("Expected a file or a directory containing `conversation.py` and `master_plan.md`.")
+                        continue
+                    backup_conversation(messages, master_plan, console)
                     messages = [{"role": i["role"], "content": i["content"]} for i in eval(readTextFile(file_path)) if i.get("role", "") in ("user", "assistant")]
                     if messages:
                         messages.insert(0, {"role": "system", "content": DEFAULT_SYSTEM})
-                    user_request = ""
-                    master_plan = ""
+                    if messages[-1].get("role", "") == "user":
+                        messages = messages[:-1]
+                    # load master plan
+                    if os.path.isdir(load_path):
+                        master_plan = readTextFile(os.path.join(load_path, "master_plan.md"))
+                        user_request = "[CONTINUE]"
+                    else:
+                        master_plan = ""
+                        user_request = ""
                     console.clear()
                     console.print(get_banner())
                     if messages:
@@ -382,14 +389,17 @@ async def main_async():
                             if i.get("role", "") in ("user", "assistant"):
                                 console.rule()
                                 console.print(Markdown(f"# {i['role']}\n\n{i['content']}"))
-                    continue
+                    if os.path.isfile(load_path) or config.agent_mode is None:
+                        # next user request
+                        continue
                 except Exception as e:
-                    pass
+                    print(f"Error: {e}\n")
+                    continue
 
             # predefined operations with `.` commands
             if user_request in action_list:
                 if user_request == ".backup":
-                    backup_conversation(console, messages, master_plan)
+                    backup_conversation(messages, master_plan, console)
                 elif user_request == ".help":
                     console.rule()
                     console.print(Markdown("Viist https://github.com/eliranwong/biblemate for help."))
@@ -413,13 +423,35 @@ async def main_async():
                 elif user_request == ".resources":
                     console.rule()
                     resources_descriptions = [f"- `//{name}`: {description}" for name, description in resources.items()]
-                    console.print(Markdown("## Available Resources\n\n"+"\n".join(resources_descriptions)))
+                    templates_descriptions = [f"- `//{name}/...`: {description}" for name, description in templates.items()]
+                    console.print(Markdown("## Available Information\n\n"+"\n".join(resources_descriptions)+"\n\n## Available Resources\n\n"+"\n".join(templates_descriptions)))
                     console.rule()
                 elif user_request == ".plans":
                     console.rule()
                     prompts_descriptions = [f"- `/{name}`: {description}" for name, description in prompts.items()]
                     console.print(Markdown("## Available Plans\n\n"+"\n".join(prompts_descriptions)))
                     console.rule()
+                elif user_request == ".edit":
+                    options = [str(i) for i in range(0, len(messages))]
+                    index_to_edit = await dialogs.getValidOptions(
+                        default=str(len(messages)-1),
+                        options=options,
+                        descriptions=[f"{messages[int(i)]['role']}: {messages[int(i)]['content'][:50]+'...' if len(messages[int(i)]['content'])>50 else messages[int(i)]['content']}" for i in options],
+                        title="Edit Conversation",
+                        text="Select an entry to edit:"
+                    )
+                    if index_to_edit:
+                        index_to_edit = int(index_to_edit)
+                        temp_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "temp", "edit.md")
+                        writeTextFile(temp_file, messages[index_to_edit]["content"])
+                        edit_file(temp_file)
+                        edited_content = readTextFile(temp_file).strip()
+                        if edited_content:
+                            messages[index_to_edit]["content"] = edited_content
+                            backup_conversation(messages, master_plan) # backup
+                            console.rule()
+                            console.print("Changes saved!", justify="center")
+                            console.rule()
                 elif user_request == ".backend":
                     edit_configurations()
                     console.rule()
@@ -455,12 +487,36 @@ async def main_async():
                     console.rule()
                     console.print("Lite Context Enabled" if config.lite else "Lite Context Disabled", justify="center")
                     console.rule()
+                elif user_request == ".download":
+                    file_ids ={
+                        "bible.db": "1E6pDKfjUMhmMWjjazrg5ZcpH1RBD8qgW",
+                        "collection.db": "1y4txzRzXTBty0aYfFgkWfz5qlHERrA17",
+                        "dictionary.db": "1UxDKGEQa7UEIJ6Ggknx13Yt8XNvo3Ld3",
+                        "encyclopedia.db": "1NLUBepvFd9UDxoGQyQ-IohmySjjeis2-",
+                        "exlb.db": "1Hpo6iLSh5KzgR6IZ-c7KuML--A3nmP1-",
+                    }
+                    file_id = await dialogs.getValidOptions(
+                        options=file_ids.keys(),
+                        title="BibleMate Data Files",
+                        text="Select a file:"
+                    )
+                    if file_id:
+                        output = os.path.join(BIBLEMATEDATA, file_id+".zip")
+                        if os.path.isfile(output):
+                            os.remove(output)
+                        if os.path.isfile(output[:-4]):
+                            os.remove(output[:-4])
+                        gdown.download(id=file_ids[file_id], output=output)
+                        with zipfile.ZipFile(output, 'r') as zip_ref:
+                            zip_ref.extractall(BIBLEMATEDATA)
+                        if os.path.isfile(output):
+                            os.remove(output)
                 elif user_request == ".mode":
                     default_ai_mode = "chat" if config.agent_mode is None else "agent" if config.agent_mode else "partner"
                     ai_mode = await dialogs.getValidOptions(
                         default=default_ai_mode,
                         options=["agent", "partner", "chat"],
-                        descriptions=["Fully automated", "Semi-automated, with review and edit prompts", "Direct text responses"],
+                        descriptions=["AGENT - Fully automated", "PARTNER - Semi-automated, with review and edit prompts", "CHAT - Direct text responses"],
                         title="AI Modes",
                         text="Select an AI mode:"
                     )
@@ -476,7 +532,7 @@ async def main_async():
                         console.print(f"`{ai_mode.capitalize()}` Mode Enabled", justify="center")
                         console.rule()
                 elif user_request in (".new", ".quit"):
-                    backup_conversation(console, messages, master_plan) # backup
+                    backup_conversation(messages, master_plan, console) # backup
                 # reset
                 if user_request == ".new":
                     user_request = ""
@@ -541,7 +597,7 @@ async def main_async():
                 console.print(Markdown(f"# User Request\n\n{user_request}\n\n# Master plan\n\n{master_plan}"))
 
             # Prompt Engineering
-            if not specified_tool == "@@" and not specified_tool == "uba" and config.prompt_engineering:
+            if not specified_tool == "@@" and config.prompt_engineering and not user_request == "[CONTINUE]":
                 async def run_prompt_engineering():
                     nonlocal user_request
                     try:
@@ -553,12 +609,8 @@ async def main_async():
                         user_request = re.sub(r"^.*?(```improved_prompt|```)(.+?)```.*?$", r"\2", user_request, flags=re.DOTALL).strip()
                 await thinking(run_prompt_engineering, "Prompt Engineering ...")
 
-            if not messages:
-                messages = [
-                    {"role": "system", "content": DEFAULT_SYSTEM},
-                    {"role": "user", "content": user_request},
-                ]
-            else:
+            # Add user request to messages
+            if not user_request == "[CONTINUE]":
                 messages.append({"role": "user", "content": user_request})
 
             async def run_tool(tool, tool_instruction):
@@ -671,9 +723,9 @@ Available tools are: {available_tools}.
             system_make_suggestion = get_system_make_suggestion(master_plan=master_plan)
 
             # Get the first suggestion
-            next_suggestion = "START"
+            next_suggestion = "CONTINUE" if user_request == "[CONTINUE]" else "START"
 
-            step = 1
+            step = int(((len(messages)-len(DEFAULT_MESSAGES)-2)/2+1)) if user_request == "[CONTINUE]" else 1
             while not ("STOP" in next_suggestion or re.sub("^[^A-Za-z]*?([A-Za-z]+?)[^A-Za-z]*?$", r"\1", next_suggestion).upper() == "STOP"):
 
                 async def make_next_suggestion():
@@ -748,6 +800,8 @@ Available tools are: {available_tools}.
 
                 await process_tool(next_tool, next_step, step_number=step)
                 console.print(Markdown(f"\n## Output [{step}]\n\n{messages[-1]['content']}"))
+                # temporaily save after each step
+                backup_conversation(messages, master_plan)
 
                 # iteration count
                 step += 1
@@ -757,7 +811,7 @@ Available tools are: {available_tools}.
                     console.rule()
                     break
 
-                # Get the next suggestion
+                # Check the progress
                 async def get_next_suggestion():
                     nonlocal next_suggestion, messages, system_progress
                     next_suggestion = agentmake([{"role": "system", "content": system_progress}]+messages[len(DEFAULT_MESSAGES):], system=system_progress, follow_up_prompt="Please decide either to `CONTINUE` or `STOP` the process.", **AGENTMAKE_CONFIG)[-1].get("content", "").strip()
@@ -785,7 +839,7 @@ Please provide me with the final answer to my original request based on the work
 
             # Backup
             print()
-            backup_conversation(console, messages, master_plan)
+            backup_conversation(messages, master_plan, console)
 
 if __name__ == "__main__":
     asyncio.run(main())
